-- This module maintains the lists of fresh indices from which to generate
-- new names and the list of captured variables

-- It is also augmented to do keep a flag for debugging and lifts IO
-- operations for debugging.

module FreshT (
  F, runF, newVar, capture, isCaptured, allCaptured, debugFlag, lift
) where

------------------------------------------------------------------------
-- The state consists of the debug flag, the next fresh index and the current
-- list of captured variables

type State = (Bool,Int,[String])

initState :: Bool -> State
initState b = (b,0,[])

newtype F a = F (State -> IO (State,a))

instance Monad F where
  return e = F (\s -> return (s,e))
  (F e1) >>= e2 = F (\s1 -> do (s2,v1) <- e1 s1
                               let F f = e2 v1
                               f s2)

runF :: Bool -> F a -> IO a
runF b (F f) = do (_,a) <- f (initState b)
                  return a

newVar :: String -> F String
newVar s = let base = takeWhile (/= '_') s
           in F (\ (b,i,ss) -> return ((b,i+1,ss), base ++ "_" ++ show i))

isCaptured :: String -> F Bool
isCaptured s = 
    F (\ (b,i,ss) -> return ((b,i,ss), s `elem` ss))

capture :: String -> F ()
capture s = 
    F (\ (b,i,ss) -> return ((b,i,s:ss), ()))

allCaptured :: F [String]
allCaptured = 
    F (\ (b,i,ss) -> return ((b,i,ss), ss))

debugFlag :: F Bool
debugFlag = 
    F (\ (b,i,ss) -> return ((b,i,ss), b))

lift :: IO a -> F a
lift c = F (\s -> do a <- c
                     return (s,a))

------------------------------------------------------------------------
