;;
;; sim_simple.ss - a simulation of call-by-need using the delimited
;; continuation framework of Dybvig, Sabry, and Peyton Jones
;;
;; sim_simple inlines all compilation.

;; Author: Ronald Garcia
;; Date: March 13, 2008

(load "control.ss")
(load "match.ss")

(print-gensym #f)


;; compile-term s t - basic compilation, mostly defers to a runtime system.
;;
;; s - the "designated" reduction prompt.
;; t - the term
(define (compile-term s t)
  (let ([kx (gensym "kx")] [ka (gensym "ka")]
        [va (gensym "va")] [th (gensym "th")]
        [f (gensym "f")] [y (gensym "y")])
    (let compile-term ([t t])
      (match t
        [,x (guard (symbol? x))
            `(withSubCont ,x
               (lambda (,kx)
                 (lambda (,th)
                   ((pushPrompt ,s (,th))
                    (lambda (,ka ,va)
                      (pushSubCont ,ka
                        ((pushPrompt ,x
                           (pushSubCont ,kx
                             (withSubCont ,s
                               (lambda (,ka)
                                 (lambda (,f) (,f ,ka ,va))))))
                         (lambda ()
                           (withSubCont ,s
                             (lambda (,ka)
                               (lambda (,f) (,f ,ka ,va))))))))))))]
      [(lambda (,x) ,[tl]) 
       `(withSubCont ,s
          (lambda (,ka)
            (lambda (,f) (,f ,ka (lambda (,x) ,tl)))))]

      [(,[t1] ,[t2])
       `((pushPrompt ,s ,t1)
         (lambda (,ka ,va) 
           (pushSubCont ,ka
             ((lambda (,y)
                ((pushPrompt ,y (,va ,y)) (lambda () ,t2)))
              (newPrompt)))))]))))


;; compile-need t - translate a lambda term into a DC-based simulation of
;; that term as run under call-by-need.  Wraps the program in the necessary
;; runtime support.
;;
;; t - the input program
(define (compile-need t)
  (let ([s (gensym)]) ;; The "designated" reduction prompt (stops "rebuild")
    `(let ([,s (newPrompt)])
       (runCC
        (lambda ()
          ((pushPrompt ,s 
             ,(compile-term s t))
           (lambda (ka va)
             (cons ka va))))))))


;; eval-need t - compile a term into a DC-based simulation and run it.  Don't
;; expect to see anything understandable when evaluation completes.
;;
;; t - the input program
(define (eval-need t) (eval (compile-need t)))
